#ifndef __IKD_CELLCONTAINER_H__
#define __IKD_CELLCONTAINER_H__


#include <stdint.h>
#include <vector>
#include <set>
#include <functional>
#include "Float2.h"


namespace Tools {

	// ZԕǗNX
	//
	//  `̋(min-max)csepYAsepXɕZԂɃIuWFNgzu
	//  Փˌms܂B
	//  IuWFNgAABBNXɕœo^܂B
	//  _̐xύXꍇFLŤ^ςĉB
	//  AABBɊ܂܂T^̃IuWFNĝ͎̏ꍇRs[p܂̂
	//  |C^X}[g|C^ŕގ߂܂B

	template< class T, class FLT = double >
	class CellContainer {
	public:
		// AABB
		//  Z̓蒷`̈\܂B
		//  AABBɊ܂܂ΏۃIuWFNg(T^)ێ܂
		class AABB {
		public:
			AABB() : minX_(), minY_(), maxX_(), maxY_(), guid_( curGuid_++ ) {}
			AABB(FLT minX, FLT minY, FLT maxX, FLT maxY, const T &obj) : minX_(minX), minY_(minY), maxX_(maxX), maxY_(maxY), guid_( curGuid_++ ) {
				checkMinMax();
				setObj( obj );
			}
			AABB(FLT minX, FLT minY, FLT maxX, FLT maxY) : minX_(minX), minY_(minY), maxX_(maxX), maxY_(maxY), guid_( curGuid_++ ) {
				checkMinMax();
			}
			AABB( const AABB &r ) {
				min_ = r.min_;
				max_ = r.max_;
				obj_ = r.obj_;
				guid_ = r.guid_;
			}
			AABB &operator =( const AABB &r ) {
				min_ = r.min_;
				max_ = r.max_;
				obj_ = r.obj_;
				guid_ = r.guid_;
				return *this;
			}

			// IuWFNgZbg
			void setObj(const T& obj) {
				obj_ = obj;
			}

			// Zbg
			void set( FLT minX, FLT minY, FLT maxX, FLT maxY , const T& obj ) {
				setAABB( minX, minY, maxX, maxY );
				setObj( obj );
			}

			// AABB̈ύX
			void setAABB( FLT minX, FLT minY, FLT maxX, FLT maxY ) {
				minX_ = minX;
				minY_ = minY;
				maxX_ = maxX;
				maxY_ = maxY;
				checkMinMax();
			}

			// AABB̈̍W擾
			const Float2Base<FLT> &getMin() const {
				return min_;
			}

			// AABB̈̉EW擾
			const Float2Base<FLT> &getMax() const {
				return max_;
			}

			// AABB̈̕ƍ擾
			Float2Base<FLT> getWH() const {
				return Flt2(maxX_ - minX_, maxY_ - minX_);
			}

			// IuWFNg擾
			const T &getObj() const {
				return obj_;
			}

			// Փ
			bool collide( const AABB &other ) const {
				return !( other.maxX_ < minX_ || other.minX_ > maxX_ || other.maxY_ < minY_ || other.minY_ > maxY_ );
			}

			// j[NId擾
			int32_t getGuid() const {
				return guid_;
			}

		private:
			// minmax̐`FbN
			void checkMinMax() {
				if (minX_ > maxX_) {
					FLT t = minX_;
					minX_ = maxX_;
					maxX_ = t;
				}
				if (minY_ > maxY_) {
					FLT t = minY_;
					minY_ = maxY_;
					maxY_ = t;
				}
			}

			union {
				struct {
					Float2Base<FLT> min_;
					Float2Base<FLT> max_;
				};
				struct {
					FLT minX_, minY_, maxX_, maxY_;
				};
			};
			T obj_;
			int32_t guid_;
			static int32_t curGuid_;
		};

		// ZCfbNXCe[^
		class CellIdxIter {
		public:
			CellIdxIter( const CellContainer *cc, const AABB &aabb ) {
				auto min = aabb.getMin();
				auto max = aabb.getMax();
				lb_ = cc->getCellIndex( min.x, min.y );
				int rb = cc->getCellIndex( max.x, min.y );
				int lt = cc->getCellIndex( min.x, max.y );
				int h = ( lt - lb_ ) / cc->sepNumX_ + 1;
				w_ = rb - lb_ + 1;
				num_ = w_ * h;
				line_ = cc->sepNumX_;
				count_ = 0;
			}

			int get() {
				if ( count_ == num_ )
					return -1;
				int idx = lb_ + count_ / w_ * line_ + count_ % w_;
				count_++;
				return idx;
			}

			bool isEnd() {
				return count_ == num_;
			}

			int lb_;
			int w_, line_;
			int num_;
			int count_;
		};

		CellContainer() : bInitialized_(), sepNumX_(), sepNumY_(), aabbNum_() {}

		// ̈
		//  AABB͈͂Zoo^s
		bool setAll( const std::vector< AABB > &aabbList, int32_t sepNumX, int32_t sepNumY ) {
			if ( aabbList.size() == 0 )
				return false;

			// ̈Zo
			Float2Base<FLT> curLb = aabbList[ 0 ].getMin();
			Float2Base<FLT> curRt = aabbList[ 0 ].getMax();
			for ( size_t i = 1; i < aabbList.size(); ++i ) {
				Float2Base<FLT> lb = aabbList[ i ].getMin();
				Float2Base<FLT> rt = aabbList[ i ].getMax();
				if ( lb.x < curLb.x )
					curLb.x = lb.x;
				if ( lb.y < curLb.y )
					curLb.y = lb.y;
				if ( rt.x > curRt.x )
					curRt.x = rt.x;
				if ( rt.y > curRt.y )
					curRt.y = rt.y;
			}
			if ( setCellRegion( curLb.x, curLb.y, curRt.x, curRt.y, sepNumX, sepNumY ) == false )
				return false;

			// AABBo^
			for ( size_t i = 0; i < aabbList.size(); ++i ) {
				addAABB( aabbList[ i ] );
			}

			return true;
		}

		// ̈
		bool setCellRegion( FLT minX, FLT minY, FLT maxX, FLT maxY, uint32_t sepNumX, uint32_t sepNumY ) {
			if (sepNumX == 0 || sepNumY == 0 || minX == maxX || minY == maxY )
				return false;

			if (minX > maxX) {
				FLT t = minX;
				minX = maxX;
				maxX = t;
			}
			if (minY > maxY) {
				FLT t = minY;
				minY = maxY;
				maxY = t;
			}
			min_ = Float2Base<FLT>(minX, minY);
			max_ = Float2Base<FLT>(maxX, maxY);
			len_ = max_ - min_;
			unit_ = Float2Base<FLT>(len_.x / sepNumX, len_.y / sepNumY);
			sepNumX_ = sepNumX;
			sepNumY_ = sepNumY;

			// Z
			for ( uint32_t i = 0; i < sepNumX * sepNumY; ++i ) {
				cells_.push_back( std::vector< AABB >() );
			}

			bInitialized_ = true;
			return true;
		}

		// AABBǉ
		int addAABB( const AABB &aabb ) {
			CellIdxIter it( this, aabb );
			while( it.isEnd() == false ) {
				int idx = it.get();
				cells_[ idx ].push_back( aabb );
			}
			return ++aabbNum_;
		}

		// Փ˂NĂIuWFNgQ擾
		bool getCollideObjects( const AABB &targetAABB, std::vector<T> &outCollideList, bool ignoreSelf = true ) {
			if ( bInitialized_ == false )
				return false;

			CellIdxIter it( this, targetAABB );
			const int32_t targetGuid = targetAABB.getGuid();
			std::set< int32_t > collidedGuid;
			while ( it.isEnd() == false ) {
				int idx = it.get();
				auto &cell = cells_[ idx ];
				for ( size_t i = 0; i < cell.size(); ++i ) {
					if ( cell[ i ].collide( targetAABB ) == true ) {
						const int32_t guid = cell[ i ].getGuid();
						if ( ignoreSelf == true && guid == targetGuid )
							continue;
						if ( collidedGuid.find( guid ) != collidedGuid.end() )
							continue;	// dՓ
						outCollideList.push_back( cell[ i ].getObj() );
						collidedGuid.insert( guid );
					}
				}
			}
			return true;
		}

		// Փ˂NĂIuWFNgQ擾
		bool getCollideObjects( const AABB &targetAABB, std::vector<AABB> &outCollideAABBList, bool ignoreSelf = true ) {
			if ( bInitialized_ == false )
				return false;

			CellIdxIter it( this, targetAABB );
			const int32_t targetGuid = targetAABB.getGuid();
			std::set< int32_t > collidedGuid;
			while ( it.isEnd() == false ) {
				int idx = it.get();
				auto &cell = cells_[ idx ];
				for ( size_t i = 0; i < cell.size(); ++i ) {
					if ( cell[ i ].collide( targetAABB ) == true ) {
						const int32_t guid = cell[ i ].getGuid();
						if ( ignoreSelf == true && guid == targetGuid )
							continue;
						if ( collidedGuid.find( guid ) != collidedGuid.end() )
							continue;	// dՓ
						outCollideAABBList.push_back( cell[ i ] );	// AABB
						collidedGuid.insert( guid );
					}
				}
			}
			return true;
		}

		// Փ˂NĂIuWFNgQCe[VĎ擾
		bool getCollideObjects( const AABB &targetAABB, bool ignoreSelf, std::function< bool( const AABB &, const AABB & ) > callback ) {
			if ( bInitialized_ == false )
				return false;

			CellIdxIter it( this, targetAABB );
			const int32_t targetGuid = targetAABB.getGuid();
			std::set< int32_t > collidedGuid;
			while ( it.isEnd() == false ) {
				int idx = it.get();
				auto &cell = cells_[ idx ];
				for ( size_t i = 0; i < cell.size(); ++i ) {
					if ( cell[ i ].collide( targetAABB ) == true ) {
						const int32_t guid = cell[ i ].getGuid();
						if ( ignoreSelf == true && guid == targetGuid )
							continue;
						if ( collidedGuid.find( guid ) != collidedGuid.end() )
							continue;	// dՓ
						// OOKorłI
						if ( callback( targetAABB, cell[ i ] ) == true )
							return true;
						collidedGuid.insert( guid );
					}
				}
			}
			return true;
		}

	private:
		// w̓_ɊYZԍ擾
		int getCellIndex( FLT x, FLT y ) const {
			if ( x < min_.x ) x = min_.x;
			else if ( x > max_.x ) x = max_.x;
			if ( y < min_.y ) y = min_.y;
			else if ( y > max_.y ) y = max_.y;
			uint32_t cx = ( int )( ( x - min_.x ) / unit_.x );
			uint32_t cy = ( int )( ( y - min_.y ) / unit_.y );
			if (cx >= sepNumX_)
				cx = sepNumX_ - 1;
			if (cy >= sepNumY_)
				cy = sepNumY_ - 1;
			return sepNumX_ * cy + cx;
		}

		bool bInitialized_;	// ς݁H
		uint32_t sepNumX_;	// X̕
		uint32_t sepNumY_;	// Y̕
		Float2Base<FLT> min_;	// ̈̍W
		Float2Base<FLT> max_;	// ̈̉EW
		Float2Base<FLT> len_;	// ̈̏c
		Float2Base<FLT> unit_;	// ̈1̏c
		uint32_t aabbNum_;	// o^ĂAABB̐
		std::vector< std::vector< AABB > > cells_;	// Z̃Xg
	};

	template< class T, class FLT > int32_t CellContainer<T, FLT>::AABB::curGuid_ = 0;

}
#endif
