#ifndef __IKD_TOOLS_JAPANMAPFACTORY_H__
#define __IKD_TOOLS_JAPANMAPFACTORY_H__


// nn}{VF[vt@C
// {ŜNp`|S_쐬l

#include <stdio.h>
#include <vector>
#include <memory>
#include <string>
#include"Shape.h"
#include "shapefiledb.h"
#include "shapepolygon.h"
#include "CellContainer.h"
#include "LinkLine.h"

namespace Tools {
	class JapanMapFactory {
	public:
		class PartData {
		public:
			size_t partsNumber_;
			Double2 min_;
			Double2 max_;
			Double2 center_;
			std::vector< Double2 > points_;
		};

		JapanMapFactory() : idxBuf_(), shapeBuf_() {}
		~JapanMapFactory() {
			delete[] idxBuf_;
			delete[] shapeBuf_;
		}

		class JapanData {
		public:
			std::string name_;
			Double2 min_;
			Double2 max_;
			Double2 center_;
			std::vector< PartData > parts_;
		};

		// {SNp`|S_쐬
		bool create() {
			// t@CVF[v𒊏o
			if ( load() == false )
				return false;

			// s{ƍsCfbNX֘At
			createPrefectureIndexMap();

			// Ss{̍sE|So
			createPoliticalBoundaryPolygons();

			// sE|S{S̃|S쐬
			std::string err = createPJapanPolygons();
			if ( err != "" ) {
				return false;
			}

			return true;
		}

		// {̒_擾
		const JapanData &getJapanData() const {
			return japanData_;
		}

	private:
		// obt@擾
		char* getBuf( const char* fileName ) {
			FILE *f = 0;
			auto err = fopen_s( &f, fileName, "rb" );
			if ( f == 0 )
				return 0;

			fseek( f, 0, SEEK_END );
			auto fileSize = ftell( f );
			fseek( f, 0, SEEK_SET );

			char *buf = new char[ fileSize ];
			fread( buf, fileSize, 1, f );

			fclose( f );

			return buf;
		}

		// VF[vt@CShape֘Awb_[[h
		bool load() {
			using namespace ShapeFile;

			const char* shxPath = "polbnda_jpn.shx";
			const char* shpPath = "polbnda_jpn.shp";

			// CfbNXi[
			idxBuf_ = getBuf( shxPath );
			if ( idxBuf_ == 0 )
				return false;
			indexHeader_ = (ShapeIndexHeader*)idxBuf_;

			// VF[vi[
			shapeBuf_ = getBuf( shpPath );
			if ( shapeBuf_ == 0 )
				return false;
			fileHeader_ = (ShapeFileHeader*)shapeBuf_;
			ShapeType shapeType = fileHeader_->getShapeType();

			// s{Ήf[^i[
			if ( db_.create( "polbnda_jpn.tsv" ) == false )
				return false;

			return true;
		}

		// s{ƍsCfbNX֘At
		void createPrefectureIndexMap() {
			for ( uint32_t i = 0; i < db_.getDataSize(); ++i ) {
				auto &data = db_.getData( i );
				if ( prefIndexMap_.find( data.pref_ ) == prefIndexMap_.end() ) {
					prefIndexMap_[ data.pref_ ] = std::vector< int >();
				}
				prefIndexMap_[ data.pref_ ].push_back( i );
			}
		}

		// Ss{̎s|So
		bool createPoliticalBoundaryPolygons() {
			int32_t recordNum = indexHeader_->getRecordNum();
			for ( int32_t i = 0; i < recordNum; ++i ) {
				ShapeFile::Polygon polygon;
				polygon.setup( fileHeader_, indexHeader_, i + 1 );
				polygons_.push_back( polygon );
			}
			return true;
		}

		// sE|S{S̃|S쐬
		std::string createPJapanPolygons() {
			std::vector< std::shared_ptr< LinkLine > > lines;	// SGbW
			std::vector< CellContainer< LinkLine* >::AABB > aabbs;			// SGbWAABBQ
			std::vector< CellContainer< LinkLine* >::AABB > outlineAABBs;	// AEgCAABBQ
			CellContainer< LinkLine* > cc;	// ʋEopZԕ
			japanData_.name_ = "Japan";
			japanData_.min_.x = japanData_.min_.y = DBL_MAX;
			japanData_.max_.x = japanData_.max_.y = DBL_MIN;

			auto pref = prefIndexMap_.begin();
			for ( ; pref != prefIndexMap_.end(); ++pref ) {
				// s{̍s
				auto regionList = pref->second;

				// @ p[c|S\钸_GbW쐬
				for ( uint32_t i = 0; i < regionList.size(); ++i ) {
					// sp[c
					CellContainer< LinkLine* > ccInRegion;	// 1sʋEopZԕ
					std::vector< std::shared_ptr< LinkLine > > srcLines;
					std::vector < CellContainer< LinkLine* >::AABB > srcAABBs;

					int32_t polygonIdx = regionList[ i ];
					auto &regionPolygon = polygons_[ polygonIdx ];
					size_t partsNum = regionPolygon.getPartsNum();
					for ( size_t j = 0; j < partsNum; ++j ) {
						auto *parts = regionPolygon.getParts( j );
						for ( int32_t p = 0; p < parts->pointNum_ - 1; ++p ) {
							srcLines.push_back( std::shared_ptr< LinkLine >( new LinkLine( parts->points_[ p ], parts->points_[ p + 1 ] ) ) );
							srcAABBs.push_back( srcLines[ srcLines.size() - 1 ]->aabb );
						}
					}
					ccInRegion.setAll( srcAABBs, 10, 10 );
					// sd菜
					std::set< int32_t > duplicateGuids;
					for ( size_t l = 0; l < srcAABBs.size(); ++l ) {
						if ( duplicateGuids.find( srcAABBs[ l ].getGuid() ) != duplicateGuids.end() ) {
							// ɏdmFς
							continue;
						}
						std::vector< CellContainer< LinkLine* >::AABB > colLines;
						ccInRegion.getCollideObjects( srcAABBs[ l ], colLines, true );
						for ( size_t cl = 0; cl < colLines.size(); ++cl ) {
							if ( srcAABBs[ l ].getObj()->isSame( colLines[ cl ].getObj(), true, true ) == true ) {
								// dXgɒǉ
								duplicateGuids.insert( colLines[ cl ].getGuid() );
							}
						}
						// ̗p
						lines.push_back( srcLines[ l ] );
						aabbs.push_back( srcLines[ l ]->aabb );

						//auto L = lines[ lines.size() - 1 ]->aabb.getObj();
						//char c[ 512 ];
						//sprintf_s( c, "%lf,%lf,%lf,%lf,", L->s_.x, L->s_.y, L->e_.x, L->e_.y );
						//OutputDebugStringA( c );
					}
				}
				//OutputDebugStringA( "\n" );
				std::string str = "JapanMap: " + pref->first + " finished.\n";
				OutputDebugStringA( str.c_str() );
			}

			// {S̍sZ̈Ƃēo^
			cc.setAll( aabbs, 25, 25 );	// SGbWAABBZԂݒ

			// A ʃGbWo菜
			for ( size_t l = 0; l < aabbs.size(); ++l ) {
				std::vector< LinkLine* > colLines;
				cc.getCollideObjects( aabbs[ l ], colLines, true );
				if ( colLines.size() == 0 ) {
					// ɐڂĂȂGbWoiG[f[^Hj
					return "Error: " + pref->first + ": No connect edge detected.";
				}
				// ʁH
				bool isDetectCommonLine = false;
				for ( size_t cl = 0; cl < colLines.size(); ++cl ) {
					if ( aabbs[ l ].getObj()->isSame( colLines[ cl ] ) == true ) {
						// ʕӌo
						isDetectCommonLine = true;

						//auto L = aabbs[ l ].getObj();
						//char c[ 512 ];
						//sprintf_s( c, "%lf,%lf,%lf,%lf,", L->s_.x, L->s_.y, L->e_.x, L->e_.y );
						//OutputDebugStringA( c );

						break;
					}
				}
				if ( isDetectCommonLine == false ) {
					// AEgCGbWێ
					outlineAABBs.push_back( aabbs[ l ] );
				}

				if ( l % 10000 == 0 ) {
					char c[ 256 ];
					sprintf_s( c, 256, "Remove common edges.  %zu / %zu\n", l, aabbs.size() );
					OutputDebugStringA( c );
				}
			}

			//OutputDebugStringA( "\n" );

			//for ( size_t l = 0; l < outlineAABBs.size(); ++l ) {
			//	auto L = outlineAABBs[ l ].getObj();
			//	char c[ 512 ];
			//	sprintf_s( c, "%lf,%lf,%lf,%lf,", L->s_.x, L->s_.y, L->e_.x, L->e_.y );
			//	OutputDebugStringA( c );
			//}
			//OutputDebugStringA( "\n" );

			// B AEgCA
			OutputDebugStringA( "Connect outlines.\n" );
			CellContainer< LinkLine* > outlineCc;	// AEgCߗ׃CopZԕ
			outlineCc.setAll( outlineAABBs, 25, 25 );
			for ( size_t l = 0; l < outlineAABBs.size(); ++l ) {
				auto obj = outlineAABBs[ l ].getObj();
				std::vector< LinkLine* > colLines;
				outlineCc.getCollideObjects( outlineAABBs[ l ], colLines, true );
				if ( colLines.size() == 0 ) {
					// ɐڂĂȂGbWoiG[f[^Hj
					return "Error: " + pref->first + ": No connect edge detected.";
				}
				bool detectLinkLine = false;
				for ( size_t cl = 0; cl < colLines.size(); ++cl ) {
					// ܂[AĂȂ̂ݑΏ
					if ( colLines[ cl ]->pre_ == 0 && obj->tryLink( colLines[ cl ], 0.0f, true ) == true ) {
						// A
						detectLinkLine = true;
						break;
					}
				}
				if ( detectLinkLine == false ) {
					// ɐڂĂȂGbWoiG[f[^Hj
					return "Error: " + pref->first + ": No connect edge detected.";
				}
			}

			// C AO[v쐬
			OutputDebugStringA( "Crete polygon groups." );
			int32_t curGroupId = 0;
			std::vector< LinkLine* > groups;
			for ( size_t l = 0; l < outlineAABBs.size(); ++l ) {
				auto *initLine = outlineAABBs[ l ].getObj();
				if ( initLine->group_ >= 0 ) {
					// ɃO[v
					continue;
				}
				initLine->group_ = curGroupId;

				// initLineɓBij܂
				// ÕCɂO[vԍtL
				LinkLine *curLine = initLine;
				bool detectUnconnect = false;
				while ( true ) {
					if ( curLine->pre_ == 0 ) {
						// ؂ĂII
						// 芸qĂOO[v
						groups.push_back( curLine );
						while ( true ) {
							curLine->group_ = curGroupId;
							if ( curLine->next_ == 0 )
								break;
							curLine = curLine->next_;
						}
						break;
					}

					curLine = curLine->pre_;

					if ( curLine == initLine ) {
						groups.push_back( initLine );
						break;
					}
					curLine->group_ = curGroupId;
				}
				curGroupId++;
			}

			// D AO[vPʂŒ_쐬
			for ( size_t g = 0; g < groups.size(); ++g ) {
				LinkLine *initLine = groups[ g ];
				LinkLine *curLine = initLine;
				PartData partData;
				partData.partsNumber_ = g;
				partData.min_ = curLine->s_, partData.max_ = curLine->s_;
				do {
					// char c[ 512 ];
					// sprintf_s( c, 512, "%f,%f,%f,%f,", curLine->s_.x, curLine->s_.y, curLine->e_.x, curLine->e_.y );
					// OutputDebugStringA( c );

					partData.points_.push_back( curLine->s_ );
					if ( curLine->s_.x < japanData_.min_.x ) japanData_.min_.x = curLine->s_.x;
					if ( curLine->s_.y < japanData_.min_.y ) japanData_.min_.y = curLine->s_.y;
					if ( curLine->s_.x > japanData_.max_.x ) japanData_.max_.x = curLine->s_.x;
					if ( curLine->s_.y > japanData_.max_.y ) japanData_.max_.y = curLine->s_.y;
					curLine = curLine->next_;
				} while ( initLine != curLine );
				partData.center_ = ( partData.min_ + partData.max_ ) / 2.0;
				japanData_.parts_.push_back( partData );
				// OutputDebugStringA( "\n" );
			}

			// {Ŝ͈̔͂vZ
			for ( size_t p = 0; p < japanData_.parts_.size(); ++p ) {
				if ( japanData_.parts_[ p ].min_.x < japanData_.min_.x ) japanData_.min_.x = japanData_.parts_[ p ].min_.x;
				if ( japanData_.parts_[ p ].min_.y < japanData_.min_.y ) japanData_.min_.y = japanData_.parts_[ p ].min_.y;
				if ( japanData_.parts_[ p ].max_.x > japanData_.max_.x ) japanData_.max_.x = japanData_.parts_[ p ].max_.x;
				if ( japanData_.parts_[ p ].max_.y > japanData_.max_.y ) japanData_.max_.y = japanData_.parts_[ p ].max_.y;
			}
			japanData_.center_ = ( japanData_.max_ + japanData_.min_ ) / 2.0;

			OutputDebugStringA( "Finish criating Japan map polygons.\n" );

			return "";
		}

		char *idxBuf_ = 0;
		char *shapeBuf_ = 0;
		ShapeFile::ShapeIndexHeader *indexHeader_;
		ShapeFile::ShapeFileHeader *fileHeader_;
		ShapeFile::ShapeFileDB db_;
		std::vector< ShapeFile::Polygon > polygons_;
		std::map< std::string, std::vector< int > > prefIndexMap_;
		JapanData japanData_;
	};
}

#endif
