#ifndef __IKD_SHAPEFILE_SHAPE_H__
#define __IKD_SHAPEFILE_SHAPE_H__

// VF[vKii.shpj

#include <stdint.h>
#include <map>

namespace ShapeFile {

	enum ShapeType {
		ShapeType_Null = 0,			// f[^
		ShapeType_Point = 1,		// 1_iXYj
		ShapeType_Polyline = 3,		// p
		ShapeType_Polygon = 5,		// p`
		ShapeType_MultiPoint = 8,	// _
		ShapeType_PointZ = 11,		// 3_iXYZAIvVMj
		ShapeType_PolylineZ = 13,	// 3p
		ShapeType_PolygonZ = 15,	// 3p`
		ShapeType_MultiPointZ = 18,	// 3_
		ShapeType_PointM = 21,		// Mt1_
		ShapeType_PolilineM = 23,	// Mtp
		ShapeType_PoligonM = 25,	// Mtp`
		ShapeType_MultiPointM = 28,	// Mt_
		ShapeType_MultiPatch = 31,	// pb`
		ShapeType_Error = 999,		// s
	};

	class ShapeFileHelper {
	public:
		// VF[v^Cv擾
		static ShapeType getShapeType( int32_t shapeTypeNum ) {
			static std::map< int32_t, ShapeType > shapeTypeMap_g = {
				{ (int)ShapeType_Null, ShapeType_Null },				// f[^
				{ (int)ShapeType_Point, ShapeType_Point },				// 1_iXYj
				{ (int)ShapeType_Polyline, ShapeType_Polyline },		// p
				{ (int)ShapeType_Polygon, ShapeType_Polygon },			// p`
				{ (int)ShapeType_MultiPoint, ShapeType_MultiPoint },	// _
				{ (int)ShapeType_PointZ, ShapeType_PointZ },			// 3_iXYZAIvVMj
				{ (int)ShapeType_PolylineZ, ShapeType_PolylineZ },		// 3p
				{ (int)ShapeType_PolygonZ, ShapeType_PolygonZ },		// 3p`
				{ (int)ShapeType_MultiPointZ, ShapeType_MultiPointZ },	// 3_
				{ (int)ShapeType_PointM, ShapeType_PointM },			// Mt1_
				{ (int)ShapeType_PolilineM, ShapeType_PolilineM },		// Mtp
				{ (int)ShapeType_PoligonM, ShapeType_PoligonM },		// Mtp`
				{ (int)ShapeType_MultiPointM, ShapeType_MultiPointM },	// Mt_
				{ (int)ShapeType_MultiPatch, ShapeType_MultiPatch },	// pb`
			};
			if ( shapeTypeMap_g.find( shapeTypeNum ) == shapeTypeMap_g.end() )
				return ShapeType_Error;
			return shapeTypeMap_g[ shapeTypeNum ];
		}

		// GfBAϊ
		static int32_t toLittle( int32_t bigEndian ) {
			return
				( ( bigEndian & 0xff ) << 24 ) |
				( ( bigEndian & 0xff00 ) << 8 ) |
				( ( bigEndian & 0xff0000 ) >> 8 ) |
				( ( bigEndian & 0xff000000 ) >> 24 );
		}
	};
	
#pragma pack(1)
	// VF[vR[h
	struct ShapeFileRecord {
		int32_t shapeType_;		// VF[v^Cv
		int32_t data_;			// f[^擪

		// f[^̐擪|C^擾
		void* getDataPtr() {
			return (void*)&data_;
		}

		// VF[v^Cv擾
		ShapeType getShapeType() {
			return ShapeFileHelper::getShapeType( shapeType_ );
		}
	};


	// VF[vR[hwb_[
	struct ShapeFileRecordHeader {
		int32_t index_big_;		// R[hCfbNX(1j
		int32_t len_big_;		// R[hi16bitWORDj

		// CfbNX擾(1j
		int32_t getIndex() {
			return ShapeFileHelper::toLittle( index_big_ );
		}

		// R[hTCY擾
		int32_t getSize() {
			return ShapeFileHelper::toLittle( len_big_ ) * 2;
		}

		// VF[vR[h̃|C^擾
		ShapeFileRecord* getRecordPtr() {
			return (ShapeFileRecord*)( this + 1 );
		}
	};

	// VF[vKit@Cwb_[
	struct ShapeFileHeader {
		int32_t id_big_;			// t@C(0x0000270a) Bigendian
		int32_t reserved_big_[ 5 ];	// gp̈
		int32_t fileLen_big_;		// t@Ci16bitWORDj
		int32_t version_;			// o[Wԍ
		int32_t shapeType_;			// Shape
		double minX_;				// ŏE`XWŏl
		double minY_;				// ŏE`YWŏl
		double maxX_;				// ŏE`XWől
		double maxY_;				// ŏE`YWől
		double minZ_;				// ŏl
		double maxZ_;				// ől
		double minM_;				// [Uŏl
		double maxM_;				// [Uől
		char recordHeader_;			// R[hwb_[擪

		// t@CTCY擾
		int32_t getFileSize() const {
			return ShapeFileHelper::toLittle( fileLen_big_ ) * 2;
		}

		// VF[v^Cv擾
		ShapeType getShapeType() {
			return ShapeFileHelper::getShapeType( shapeType_ );
		}

		// R[hwb_[擾	
		ShapeFileRecordHeader* getRecordHeaderPtr( int idx = 0 ) const {
			ShapeFileRecordHeader* ary = (ShapeFileRecordHeader*)&recordHeader_;
			return ary + idx;
		}
	};

	// VFCvCfbNX
	struct ShapeIndex {
		int32_t recordPos_big_;		// R[hʒu
		int32_t recordLen_big_;		// R[h(16bitWORD)

		// R[hTCY擾
		int getSize() {
			return ShapeFileHelper::toLittle( recordLen_big_ ) * 2;
		}

		// R[hʒu擾
		int getPos() {
			return ShapeFileHelper::toLittle( recordPos_big_ ) * 2;
		}
	};

	// VFCvCfbNXwb_[
	struct ShapeIndexHeader : ShapeFileHeader {
		// CfbNXwb_[擾	
		ShapeIndex* getShapeIndexPtr( int idx ) const {
			ShapeIndex* ptr = (ShapeIndex*)&recordHeader_;
			return ptr + ( idx - 1 );
		}

		// R[h擾
		int32_t getRecordNum() const {
			int32_t fileSize = getFileSize();
			return ( fileSize - 100 ) / sizeof( ShapeIndex );
		}
	};


#pragma pack()
}

#endif
