#ifndef __IKD_SHAPEFILE_POLYGON_H__
#define __IKD_SHAPEFILE_POLYGON_H__

#include "Float2.h"
#include "Shape.h"
#include <vector>

namespace ShapeFile {

	using namespace Tools;

#pragma pack(1)
	struct AABB {
		Double2 min_;
		Double2 max_;

		AABB() : min_( DBL_MAX, DBL_MAX ), max_( DBL_MIN, DBL_MIN ) {}
		void setPoint( const Double2 &p ) {
			if ( p.x < min_.x ) min_.x = p.x;
			if ( p.y < min_.y ) min_.y = p.y;
			if ( p.x > max_.x ) max_.x = p.x;
			if ( p.y > max_.y ) max_.y = p.y;
		}
	};

	struct PolygonDesc {
		double minX_;
		double minY_;
		double maxX_;
		double maxY_;
		int32_t partsNum_;
		int32_t pointNum_;

		// p[c擪CfbNXz̐擪|C^擾
		int32_t* getPartsIdxPtr() const {
			return (int32_t*)( this + 1 );
		}

		// XYWz̐擪|C^擾
		Double2* getCoordPtr() const {
			return (Double2*)((int32_t*)(this + 1) + partsNum_ );
		}
	};
#pragma pack()

	// p[c
	class Parts {
	public:
		Parts() {}
		bool setup( const PolygonDesc *desc, int32_t partsIndex ) {
			index_ = partsIndex;
			if ( partsIndex >= desc->partsNum_ )
				return false;
			if ( partsIndex == desc->partsNum_ - 1 ) {
				// Ō̃p[cȂ̂œ_̐̓|CgtZ
				pointNum_ = desc->pointNum_ - desc->getPartsIdxPtr()[ index_ ];
			}
			else {
				// ̃p[c܂ł|Cg
				int32_t *p = desc->getPartsIdxPtr();
				pointNum_ = p[ index_ + 1 ] - p[ index_ ];
			}
			points_ = desc->getCoordPtr() + index_;

			// ψʒuAABBZo
			for ( int32_t i = 0; i < pointNum_; ++i ) {
				avePos_ += points_[ i ];
				aabb_.setPoint( points_[ i ] );
			}
			avePos_ /= pointNum_;

			return true;
		}

		const Double2& getAverage() const {
			return avePos_;
		}

		const AABB& getAABB() const {
			return aabb_;
		}

		int32_t index_;
		int32_t pointNum_;
		Double2* points_;
		Double2 avePos_;
		AABB aabb_;
	};

	// |S
	class Polygon {
	public:
		Polygon() {

		}
		~Polygon() {

		}

		// |Sݒ
		bool setup( const ShapeFileHeader *header, const ShapeIndexHeader *idxHeader, int32_t recordIdx ) {
			header_ = header;
			idxHeader_ = idxHeader;

			if ( recordIdx - 1 >= idxHeader_->getRecordNum() ) {
				return false;
			}
			index_ = recordIdx;

			// f[^̊i[
			auto shapeIdx = idxHeader->getShapeIndexPtr( index_ );
			int32_t pos = shapeIdx->getPos();
			ShapeFileRecordHeader *record = ( ShapeFileRecordHeader*)( (char*)header + pos );
			ShapeFileRecord *data = record->getRecordPtr();
			desc_ = (PolygonDesc*)data->getDataPtr();

			// p[ci[
			for ( int32_t i = 0; i < desc_->partsNum_; ++i ) {
				Parts p;
				p.setup( desc_, i );
				partsList_.push_back( p );
			}
			return true;
		}

		// p[c擾
		size_t getPartsNum() const {
			return partsList_.size();
		}

		// p[c擾
		const Parts *getParts( int32_t partsIdx ) const {
			if ( partsIdx >= (int32_t)partsList_.size() )
				return 0;
			return &partsList_[ 0 ] + partsIdx;
		}

	private:
		int32_t index_;	// R[hCfbNX
		const ShapeFileHeader *header_;
		const ShapeIndexHeader *idxHeader_;
		PolygonDesc *desc_;
		std::vector< Parts > partsList_;
	};
}
#endif
